// rtos.c
// 2018.11.20
// RTOS nach DI_HALT http://easyelectronics.ru
// benutzt typedef aus mein_def.h, Ueberschaubarkeit optimisiert
// Ringpuffer fuer TaskQueue
// die Aufgaben bekommen je 3 Parameter: u16,s16,u8
#include <avr/io.h>
#include <avr/interrupt.h>

#include "rtos.h"

/********************************************************/
/*** Fuer Parameter Uebergabe an die Aufgaben ***********/
/********************************************************/
struct ParAufgabe PA_puff; // Parameter-Puffer fuer Aufgaben

/********************************************************/
/***  Fuer Ringpuffer TaskQueue *************************/
/********************************************************/

static u8 task_count = 0;
static u8 task_schwanz = 0;
static u8 task_kopf = 0;

static void task_count_inc(u8 *count, u8 max_wert){
	(*count)++;
	if(*count >= max_wert){
			*count = 0;
		}
}

/********************************************************/


/********************************************************/
/*************** RTOS ***********************************/
/*** Reihen von Aufgaben und Timer **********************/
/*** Datentyp = Zeiger auf Funktion *********************/
/********************************************************/
//volatile static TPTR	TaskQueue[TaskQueueSize+1];		// Reihe von Zeiger

volatile static struct 	{									
						TPTR Task; 					// Zeiger auf Funktion
						u16 par0;				// Parameter 0
						s16 par1;				// Parameter 1
						u8  par2;				// Parameter 2
						} TaskQueue[TaskQueueSize+1];	// Reihe von Zeiger

volatile static struct 	{									
						TPTR GoToTask; 					// Zeiger auf Funktion
						u16 Time;				// Verzoegerung, ms
						u16 par0;				// Parameter 0
						s16 par1;				// Parameter 1
						u8  par2;				// Parameter 2
						} MainTimer[MainTimerQueueSize+1];	// Reihe von Timer

/****************************************************************************************/
/*** Leerlauf ***************************************************************************/
/****************************************************************************************/
inline void Idle(void){

}

/****************************************************************************************/
/***** RTOS Vorbereitung. Reihen auf Null ***********************************************/
/****************************************************************************************/
inline void InitRTOS(void)
{
u8	index;

for(index=0;index!=TaskQueueSize+1;index++)	// Ueberall schreiben Idle
	{
	TaskQueue[index].Task = Idle;
	TaskQueue[index].par0 = 0;
	TaskQueue[index].par1 = 0;
	TaskQueue[index].par2 = 0;
	}


for(index=0;index!=MainTimerQueueSize+1;index++) // Alle Timer > 0
	{
	MainTimer[index].GoToTask = Idle;
	MainTimer[index].Time = 0;
	MainTimer[index].par0 = 0;
	MainTimer[index].par1 = 0;
	MainTimer[index].par2 = 0;
	}

	PA_puff.par0 = 0;
	PA_puff.par1 = 0;
	PA_puff.par2 = 0;
}

/*******************************************************************************************/
/* Funktion, um eine Aufgabe in die Reihe zu stellen. Zeiger auf Funktion wird uebergeben. */
/*******************************************************************************************/
void SetTask(TPTR TS,u16 par0, s16 par1, u8 par2){

	u8 tmp_sreg;	// temporaerer Speicher fuer das Statusregister
  	tmp_sreg = SREG; // Statusregister (also auch das I-Flag darin) sichern
  	cli();                                          // read and clear atomic !

	// task_kopf zeigt auf naechsten freien Platz
	// task_schwanz zeigt auf die am fruehesten gestellte Aufgabe
	// task_count zeigt die Aufgabenzahl in dem Puffer

	if(task_count < TaskQueueSize){	
	// Wenn puffer noch Platz hat

		TaskQueue[task_kopf].Task = TS;
		TaskQueue[task_kopf].par0 = par0;
		TaskQueue[task_kopf].par1 = par1;
		TaskQueue[task_kopf].par2 = par2;
		task_count++;

		task_count_inc(&task_kopf, TaskQueueSize);
	}

  	SREG = tmp_sreg; // Status-Register wieder herstellen
}

/********************************************************************************************/
/* Funktion, die eine Aufgabe in die Timerreihe stellt. Zeiger auf Funktion wird uebergeben.*/
/***** Zeit in Einheiten von Systemtimer. normalerweise 1 ms ********************************/
/********************************************************************************************/
void SetTimerTask(TPTR TS, u16 NewTime,u16 par0, s16 par1, u8 par2){
	u8 index=0;

	u8 tmp_sreg;	// temporaerer Speicher fuer das Statusregister
  	tmp_sreg = SREG; // Statusregister (also auch das I-Flag darin) sichern
  	cli();                                          // read and clear atomic !


for(index=0;index!=MainTimerQueueSize+1;++index){	// Suche in der Timerreihe
	if(MainTimer[index].GoToTask == TS){			// Wenn gleiche Aufgabe gefunden wird
		MainTimer[index].Time = NewTime;		// neue Zeitangabe aufschreiben
		MainTimer[index].par0 = par0;		// Aufgaben in Puffer
		MainTimer[index].par1 = par1;
		MainTimer[index].par2 = par2;
		SREG = tmp_sreg; // Status-Register wieder herstellen
		return;						// Hier koennte die Kode fur erfolgreiche Suche zurueckgegeben werden
		}
	}
	

for(index=0;index!=MainTimerQueueSize+1;++index){	// Wenn keine gleiche Aufgabe gefunden wurde, Suche nach dem freien Timer
	if(MainTimer[index].GoToTask == Idle){
		MainTimer[index].GoToTask = TS;			// Zeiger auf Aufgabe aufschreiben
		MainTimer[index].Time = NewTime;		// Zeit aufschreiben
		MainTimer[index].par0 = par0;		// Aufgaben in Puffer
		MainTimer[index].par1 = par1;
		MainTimer[index].par2 = par2;
		SREG = tmp_sreg; // Status-Register wieder herstellen
		return;				//  
		}
	}	// hier koennte Fehlerkode zurueckgegeben werden: keine freien Timer
}

/****************************************************************************************/
/*** Aufgaben-Dispatcher. Nimmt die Aufgaben aus der Aufgabenreihe und erfuellt sie *****/
/****************************************************************************************/

inline void TaskManager(void){

	TPTR GoToTask = Idle;	// Variablen initialisieren

	cli();		// Interrupts verbieten!!!

	// task_kopf zeigt auf naechsten freien Platz
	// task_schwanz zeigt auf die am fruehesten gestellte Aufgabe
	// task_count zeigt die Aufgabenzahl in dem Puffer

	if(task_count){
		GoToTask = TaskQueue[task_schwanz].Task;	// Ersten Wert aus der Aufgabenreihe nehmen.
		PA_puff.par0 = TaskQueue[task_schwanz].par0;
		PA_puff.par1 = TaskQueue[task_schwanz].par1;
		PA_puff.par2 = TaskQueue[task_schwanz].par2;
		task_count--;
		task_count_inc(&task_schwanz, TaskQueueSize);
		sei();							// Interrupts erlauben
		(GoToTask)();					// Aufgabe erfuellen
	}else{
		sei();					// Interrupts erlauben
		(Idle)(); 				// Idle bearbeiten
	}
}

/****************************************************************************************/
/***** Timer-Dienst. Sollte jede 1 ms in Interrupt gemacht werden. **********************/
/***** Die Zeit kann auch anders werden, je nach der Aufgabe ****************************/
/****************************************************************************************/
inline void TimerService(void){
	u8 index;

	for(index=0;index!=MainTimerQueueSize+1;index++){	// Reihe von Timer ueberpruefen
		if(MainTimer[index].GoToTask == Idle) continue;	// wenn leer, weiter gehen

		if(MainTimer[index].Time !=1){	// wenn noch nicht leer, Decrement. 

			MainTimer[index].Time --;		// 
		}else{
			SetTask(MainTimer[index].GoToTask,MainTimer[index].par0,MainTimer[index].par1,MainTimer[index].par2);	// wenn 0 geworden, die Aufgabe in die Aufgabenreihe stellen
			MainTimer[index].GoToTask = Idle;	// und in Timer-Celle Idle aufschreiben
		}
	}
}

//====================================================== 
